//
//  main.c
//
//  Solar-cell Charge Controller 0.2c
//  Copyright (C) 2009 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Solar-cell to lead-acid battery chager and load limmiter.
//		Charging method is 2-phase CC/CV(constant current & constant voltage) 
//		Full charge after in the trickle charge mode.
//		If it under solar power condition, the MPPT control.
//		Circuit type is buck+boost DC-DC converter. (V0.2c)
//
//	SolarCell:	VOC_SUNSET~22V (R=1:10 FVR=2V)
//	Battery:	Lead-acid type ~22V (R=1:10 FVR=2V)
//	Load:		(Isol + Ibat + Iload*2) < 10A (R=0.1 FVR=1V)
//
//  Device: PIC18F14K50
//
//  Pin assign:
//		p1:  Vdd
//		p2:  RA5      
//		p3:  RA4 AN3  
//		p4:  RA3 MCLR <-- SW1
//		p5:  RC5 CCP  --> Boost gate
//		p6:  RC4 CCP  --> Buck gate
//		p7:  RC3 AN7  <-- Isc
//		p8:  RC6 AN8  <-- Icharge
//		p9:  RC7      <-- VBUS sense
//		p10: RB7      --> Load gate
//		p11: RB6 SCL  
//		p12: RB5 AN11 <-- Iload
//		p13: RB4 SDA  
//		p14: RC2 AN6  <-- Vbat
//		p15: RC1 AN5  <-- Vsol
//		p16: RC0 AN4  <-- Vref(opt)
//		p17: VUSB
//		p18: RA1 D-
//		p19: RA0 D+
//		p20: Vss
//
//	Clock:	12MHz/Xtal -> Fosc=6MHz
//
//	Internal peripheral module:
//		WDT:	4sec
//		BOR:	-
//		EEPROM:	-
//		Timer0:	Interval timer
//		Timer1:	-
//		Timer2:	PWM
//		Timer3:	-
//		Comp:	-
//		ECCP1:	PWM
//		USB:	Serial emuration(low speed)
//		MSSP:	-
//		EUSART:	-
//		ADC:	Voltage/Current measure
//		FVR:	Use ADC
//
//	Environment:
//		IDE: MPLAB IDE 8.36
//      C Compiler for PIC18 MCUs v3.33(need -Oi+ option)
//      Microchip Application Libraries v2009-07-10
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2009.09.14 v0.2  First cording
//    2009.10.06 v0.2c Change LCD to USB serial I/F
//    2009.10.18 v0.2c Vref V/R -> FVR(internal)
//

#include <stdlib.h>
#include <string.h>

#include "p18f14K50.h"

#include "adc.h"						// Microchip C18 library
// vvv  Bug?   adc.h:550~552
#define ADC_REF_VDD_VDDb		0b11110011 	// ADC voltage source VREF+ = AVDD
#define ADC_REF_VDD_VREFPLUSb 	0b11110111 	// ADC voltage source VREF+ = ext. source at VREF+
#define ADC_REF_VDD_FVREFb		0b11111011 	// ADC voltage source VREF+ = FVREF+
#include "delays.h"						// Microchip C18 library
#include "timers.h"						// Microchip C18 library
#include "pwm.h"						// Microchip C18 library

#include "GenericTypeDefs.h"			// Microchip Application Libraries v2009-07-10
#include "Compiler.h"					// Microchip Application Libraries v2009-07-10
#include "usb_config.h"					// Microchip Application Libraries v2009-07-10
#include "USB/usb_device.h"				// Microchip Application Libraries v2009-07-10
#include "USB/usb.h"					// Microchip Application Libraries v2009-07-10
#include "USB/usb_function_cdc.h"		// Microchip Application Libraries v2009-07-10

void itoanm(int i, char *a, char n, char m);	// Private sub

//
// Application constant
//
#define CSV_HEAD						// USB output data
#define STAT_DELAY		20				// USB connect -> statistics output delay sec
#define CLIMBSRCH		2				// duty control method (down step count)
#define VOC_SUNSET		1000			// To night threthold[mV]
#define VOC_SUNRISE		4000			// To day threthold[mV]
#define VOC_BOOSTON		16000			// Boost start threthold[mV]
#define VOC_BUCKON		16500			// Buck start threthold[mV]
#define VBAT_MAX		14700			// 1st limit[mV]
#define VBAT_TRICKLE	13700			// 2nd limit[mV]
#define VBAT_DECR		13000			// Revival to max charge[mV]
#define VBAT_NORMAL		12500			// Revival to load on[mV]
#define VBAT_EMPTY		11500			// Low limit, load off[mV]
#define IBAT_MAX		1500			// Charge Normal 0.1C[mA]
#define IBAT_CHGOFF		1000			// Down slope cutoff MAX2/3[mA]
#define IBAT_TRICKLE	150				// Charge Little 0.01C[mA]
#define PWMPERIOD		159				// 37.5KHz / 6MHz
#define PWMHILIM_BUCK	158				// duty high limit
#define PWMLOLIM_BUCK	1				// duty low limit
#define PWMSTEP_BUCK	2				// duty step
#define PWMHILIM_BOOST	39				// duty high limit
#define PWMLOLIM_BOOST	1				// duty low limit
#define PWMSTEP_BOOST	1				// duty step
#define PWMHILIM_THROUGH PWMPERIOD		// duty high limit
#define PWMLOLIM_THROUGH 1				// duty low limit
#define PWMSTEP_THROUGH  4				// duty step
// Hardware
#define FVR_ON							// Vref=FVR enable
#define AD_VDD  		2700			// [mV]
#define AD_RES  		1024			// resolution 10bit
#define DIVREG1 		111				// Vsol Need adjust
#define DIVREG2 		10				//      11.1:1
#define DIVREG3 		111				// Vbat Need adjust
#define DIVREG4			10				//      11.1:1
#define SENSREG1		10				// Isol    1/10[ohm]
#define SENSREG2		10				// Icharge 1/10[ohm]
#define SENSREG3		10				// Iload   1/10[ohm]
#define TMR0VAL			-5859			// Interval = 1sec < CLK(6MHz) / 4 / PS(256) >
// ID
#define WORLD_NIGHT		0				// world_stat
#define WORLD_DAY		1				// 
#define CHG_OFF			0				// charge_mode
#define CHG_THROUGH		1				//
#define CHG_BOOST		2				//
#define CHG_BUCK		3				//

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define ABS(a)		(((a) >  0 ) ? (a) : -(a))

//
// Configuration bits
//
#pragma config CPUDIV = CLKDIV2			// 6MHz
#pragma config USBDIV = ON				// Low speed
#pragma config FOSC   = HS				// Xtal=12MHz
#pragma config PLLEN  = OFF				// Power reduce
#pragma config FCMEN  = OFF
#pragma config IESO   = OFF
#pragma config PWRTEN = OFF
#pragma config BOREN  = OFF
#pragma config BORV   = 19
//#pragma config VREGEN = ON
#pragma config WDTEN  = OFF
#pragma config WDTPS  = 1024			// 4sec
#pragma config MCLRE  = OFF
#pragma config HFOFST = OFF
#pragma config STVREN = ON
#pragma config LVP    = OFF
#pragma config XINST  = OFF
#pragma config BBSIZ  = OFF
#pragma config CP0    = OFF
#pragma config CP1    = OFF
#pragma config CPB    = OFF
#pragma config WRT0   = OFF
#pragma config WRT1   = OFF
#pragma config WRTB   = OFF
#pragma config WRTC   = OFF
#pragma config EBTR0  = OFF
#pragma config EBTR1  = OFF
#pragma config EBTRB  = OFF       

//
// Global area
//
// Inter function parameter
int vref;								// [mV]
int vsoloc;								// [mV]
int vsol;								// [mV]
int vbat;								// [mV]
int isol;								// [mA]
int ibat;								// [mA]
int ilod;								// [mA]
int psol;								// [mW]
int pbat;								// [mW]
int plod;								// [mW]
int tic = 0;							// Timer overflow count

//
// Private work area
//
int  duty     = 0;						// PWM duty rate
int  incr     = PWMSTEP_BOOST;			// next duty offset
int  ibat_b   = 0;						// before ibat
int  vol_lim  = VBAT_MAX;				//
int  cur_lim  = IBAT_MAX;				//
char climbctr = CLIMBSRCH;				// Up hill counter

//-----------------------------------------------------------------------------

//
// Interrupt Service Routine
//
#pragma code
#pragma interrupt HighPriorityISRCode
void HighPriorityISRCode() {
	if (INTCONbits.TMR0IE && INTCONbits.TMR0IF) {
		WriteTimer0(TMR0VAL);
		INTCONbits.TMR0IF = 0;
		tic++;
	} else {
		USBDeviceTasks();
	}
}
#pragma interruptlow LowPriorityISRCode
	void LowPriorityISRCode() {
}

#define REMAPPED_RESET_VECTOR_ADDRESS			0x00
#define REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS	0x08
#define REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS	0x18
#pragma code REMAPPED_HIGH_INTERRUPT_VECTOR = REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS
void Remapped_High_ISR (void) {
     _asm goto HighPriorityISRCode _endasm
}
#pragma code REMAPPED_LOW_INTERRUPT_VECTOR = REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS
void Remapped_Low_ISR (void) {
     _asm goto LowPriorityISRCode _endasm
}
#pragma code

//
// USB Device support (Microchip library)
//
// Callbacks
void USBCBSuspend(void) {
}

void USBCBWakeFromSuspend(void) {
}

void USBCB_SOF_Handler(void) {
}

void USBCBErrorHandler(void) {
}

void USBCBCheckOtherReq(void) {
    USBCheckCDCRequest();
}

void USBCBStdSetDscHandler(void) {
}

void USBCBInitEP(void) {
    CDCInitEP();
}

void USBCBSendResume(void) {
    static WORD delay_count;
    
    USBResumeControl = 1;                // Start RESUME signaling
    delay_count = 1800U;                // Set RESUME line for 1-13 ms
    do {
        delay_count--;
    } while(delay_count);
    USBResumeControl = 0;
}

#if defined(ENABLE_EP0_DATA_RECEIVED_CALLBACK)
void USBCBEP0DataReceived(void) {
}
#endif

BOOL USER_USB_CALLBACK_EVENT_HANDLER(USB_EVENT event, void *pdata, WORD size) {
    switch(event) {
        case EVENT_CONFIGURED: 
            USBCBInitEP();
            break;
        case EVENT_SET_DESCRIPTOR:
            USBCBStdSetDscHandler();
            break;
        case EVENT_EP0_REQUEST:
            USBCBCheckOtherReq();
            break;
        case EVENT_SOF:
            USBCB_SOF_Handler();
            break;
        case EVENT_SUSPEND:
            USBCBSuspend();
            break;
        case EVENT_RESUME:
            USBCBWakeFromSuspend();
            break;
        case EVENT_BUS_ERROR:
            USBCBErrorHandler();
            break;
        case EVENT_TRANSFER:
            Nop();
            break;
        default:
            break;
    }      
    return TRUE; 
}

// Middle application program interface
void tputs(char *s) {
	if ((USBDeviceState >= CONFIGURED_STATE) && (USBSuspendControl != 1)) {
		int i;

		for(i = 1000; i > 0; --i) {
			if(mUSBUSARTIsTxTrfReady()) break;
			CDCTxService();
		}
		if (i > 0) {
			char *pData;
			BYTE len;

#ifndef AAA
 			putsUSBUSART(s);
#else
			// *** Copy from putsUSBUSART ***  bug?(output 0x00 data extra)
			USBMaskInterrupts();
			if(cdc_trf_state != CDC_TX_READY){
				USBUnmaskInterrupts();
				return;
			}
			len = 0;
			pData = s;
			while(*pData++) {
				if (++len == 255) break;
			}
			mUSBUSARTTxRam((BYTE*)s, len);
			USBUnmaskInterrupts();
			// *** Copy from putsUSBUSART ***
#endif

			CDCTxService();
		}
	}
	return;
}
#ifdef CSV_HEAD
void tputrs(const ROM char *s) {
	if ((USBDeviceState >= CONFIGURED_STATE) && (USBSuspendControl != 1)) {
		int i;

		for(i = 1000; i > 0; --i) {
			if(mUSBUSARTIsTxTrfReady()) break;
			CDCTxService();
		}
		if (i > 0) {
 			putrsUSBUSART(s);
			CDCTxService();
		}
	}
	return;
}
#endif

//********************************************************************
//	A / D  c o n v e r t
//********************************************************************
int getadc(unsigned char ch, unsigned char ref) {
	int rslt;
//	char i;

//#define GETADC_AVG 1
	OpenADC(ADC_FOSC_8 &				// clock<-6MHz
			ADC_RIGHT_JUST &
			ADC_0_TAD, 
			ADC_INT_OFF &
			ch,
			ref,
			0x09f0);
//	rslt = 0;
//	for(i = 0; i < GETADC_AVG; i++) {
//		Delay10TCYx(10);
		ConvertADC();
		while(BusyADC());
		rslt = ReadADC();
//		rslt += ReadADC();
//	}
//	rslt /= GETADC_AVG;
	CloseADC();
	return(rslt);
}

//********************************************************************
//	M e a s u r e
//********************************************************************
void measure(unsigned char oc) {	// option: get vsoloc (Slowly)
	int avref;
	int adc_a;
	int adc_b;
	int adc_1;
	int adc_2;
	int adc_3;
	int v1;
	int v2;


#ifdef FVR_ON
	// Pin read
#define VREF1 2048
	REFCON0bits.FVR1S0 = 0;				// 2.048V
	REFCON0bits.FVR1S1 = 1;
	REFCON0bits.FVR1EN = 1;
	while(!REFCON0bits.FVR1ST);
	adc_a = getadc(ADC_CH5,  ADC_REF_VDD_FVREFb & ADC_REF_VDD_VSS);
	adc_b = getadc(ADC_CH6,  ADC_REF_VDD_FVREFb & ADC_REF_VDD_VSS);
	REFCON0bits.FVR1EN = 0;

#define VREF2 1024
	REFCON0bits.FVR1S0 = 1;				// 1.024V
	REFCON0bits.FVR1S1 = 0;
	REFCON0bits.FVR1EN = 1;
	while(!REFCON0bits.FVR1ST);
	adc_1 = getadc(ADC_CH7,  ADC_REF_VDD_FVREFb & ADC_REF_VDD_VSS);
	adc_2 = getadc(ADC_CH8,  ADC_REF_VDD_FVREFb & ADC_REF_VDD_VSS);
	adc_3 = getadc(ADC_CH11, ADC_REF_VDD_FVREFb & ADC_REF_VDD_VSS);
	REFCON0bits.FVR1EN = 0;

	// Convert  mV,mA
	v1   = (long)adc_1         * VREF2                     / AD_RES;
	v2   = (long)adc_2         * VREF2                     / AD_RES;
	vsol = (long)adc_a         * VREF1 * DIVREG1 / DIVREG2 / AD_RES - v1;
	vbat = (long)adc_b         * VREF1 * DIVREG3 / DIVREG4 / AD_RES - v2;
	isol = (long)adc_1         * VREF2 * SENSREG1          / AD_RES;
	ilod = (long)(adc_3-adc_2) * VREF2 * SENSREG3          / AD_RES;
	ibat = (long)(adc_2-adc_1) * VREF2 * SENSREG2          / AD_RES - ilod;

#else
	// Pin read
	avref = getadc(ADC_CH4,  ADC_REF_VDD_VDDb      & ADC_REF_VDD_VSS);
	adc_a = getadc(ADC_CH5,  ADC_REF_VDD_VDDb      & ADC_REF_VDD_VSS);
	adc_b = getadc(ADC_CH6,  ADC_REF_VDD_VDDb      & ADC_REF_VDD_VSS);
	adc_1 = getadc(ADC_CH7,  ADC_REF_VDD_VREFPLUSb & ADC_REF_VDD_VSS);
	adc_2 = getadc(ADC_CH8,  ADC_REF_VDD_VREFPLUSb & ADC_REF_VDD_VSS);
	adc_3 = getadc(ADC_CH11, ADC_REF_VDD_VREFPLUSb & ADC_REF_VDD_VSS);

	// Convert  mV,mA
	vref = (long)avref         * AD_VDD                     / AD_RES;
	v1   = (long)adc_1         * vref                       / AD_RES;
	v2   = (long)adc_2         * vref                       / AD_RES;
	vsol = (long)adc_a         * AD_VDD * DIVREG1 / DIVREG2 / AD_RES - v1;
	vbat = (long)adc_b         * AD_VDD * DIVREG3 / DIVREG4 / AD_RES - v2;
	isol = (long)adc_1         * vref   * SENSREG1          / AD_RES;
	ilod = (long)(adc_3-adc_2) * vref   * SENSREG3          / AD_RES;
	ibat = (long)(adc_2-adc_1) * vref   * SENSREG2          / AD_RES - ilod;

#endif

	// Adjust                             *** Need modify ***
//	ibat  =  (long)ibat * 15 / 16;			// *93.8%
	vbat -= ((long)ibat * 15 / 100);		// -(0.15 * ibat)
	vbat += 170;							// +0.17[V]
	if (isol != 0) {
		isol += 50;							// +0.05[A]
	} 

	// Calc Watt
	psol = (long)vsol * isol / 1000;
	pbat = (long)vbat * ibat / 1000;
	plod = (long)vbat * ilod / 1000;

	// Get Open Circuit voltage
	if (oc) {
		unsigned char s1;
		unsigned char s2;

		// Save port status
		s1 = LATC;
		s2 = PSTRCON;

		// Solar-cell disconnect
		LATCbits.LATC4   = 0;		// Close buck gate
		PSTRCONbits.STRB = 0;		// Disconnect buck gate from PWM
		PSTRCONbits.STRA = 0;		// Disconnect boost gate from PWM

		// Pin read
		Delay10KTCYx(4);			// 20ms~: stable wait solar-cell 8MHz
		adc_a = getadc(ADC_CH5, ADC_REF_VDD_VDDb & ADC_REF_VDD_VSS);

		// Resume port
		PSTRCON = s2;
		LATC    = s1;

		// Convert  mV
		vsoloc = (long)adc_a * AD_VDD * DIVREG1 / DIVREG2 / AD_RES;
	}
	return;
}

//********************************************************************
//	C h a r g e   c o n t r o l  -  Boost
//********************************************************************
unsigned char chg_boost(unsigned char mode) {
//	static int  duty     = 0;			// PWM duty rate
//	static int  incr     = PWMSTEP_BOOST;// next duty offset
//	static int  ibat_b   = 0;			// before ibat
//	static int  vol_lim  = VBAT_MAX;	//
//	static int  cur_lim  = IBAT_MAX;	//
//	static char climbctr = CLIMBSRCH;	// Up hill counter


	//
	// In boost mode
	//
	if (mode == CHG_BOOST) {
		// PWM module open
		if (!PSTRCONbits.STRA) {		// PWM not set
			OpenTimer2(TIMER_INT_OFF & T2_PS_1_1 & T2_POST_1_1);
			OpenPWM1(PWMPERIOD >> 2);	// Divide 4
			LATCbits.LATC4   = 1;		// Enable buck gate
			LATCbits.LATC5   = 0;		// Disable boost gate
			PSTRCONbits.STRB = 0;		// Disconnect buck gate from PWM
			PSTRCONbits.STRA = 1;		// Connect boost gate to PWM
			duty = PWMLOLIM_BOOST;
		}

		// Taget limit
		if (vol_lim >= VBAT_MAX) {		// Max charge mode
			if ((vbat >= VBAT_MAX) && (ibat < IBAT_CHGOFF)) {
				vol_lim = VBAT_TRICKLE;	// End of max charge
				cur_lim = IBAT_TRICKLE;
			}
		} else { 						// Trickle(float) charge mode
			if (vbat < VBAT_DECR) {
				vol_lim = VBAT_MAX;		// Start max charge
				cur_lim = IBAT_MAX;
			}
		}

		// Target seek
		if ((vbat > vol_lim) || (ibat > cur_lim)) {
			duty -= PWMSTEP_BOOST;
			if (duty < PWMLOLIM_BOOST) {
				duty = PWMLOLIM_BOOST;
				mode = CHG_OFF;			// Fail! Force finish
			}
		} else {
			int d;

			d = ibat - ibat_b;
			if (d < 0) {				// Down? Continue?
				if (--climbctr <= 0) {
					incr = -incr;
					climbctr = CLIMBSRCH;
				}
			} else if (d > 0) {
				climbctr = CLIMBSRCH;
			}
		    duty  += incr;
			ibat_b = ibat;
			if (duty < PWMLOLIM_BOOST) {
				duty = PWMLOLIM_BOOST;
				incr = -incr;
			} else if (duty > PWMHILIM_BOOST) {
				duty = PWMHILIM_BOOST;
				incr = -incr;
			}
		}
		SetDCPWM1(duty);

	//
	// Fade out boost mode
	//
	} else {
		if (duty > PWMLOLIM_BOOST) {
			duty -= PWMSTEP_BOOST;
			SetDCPWM1(duty);
			mode = CHG_BOOST;			// Continued
		} else {
			LATCbits.LATC4 = 0;			// Disable buck gate
			LATCbits.LATC5 = 0;			// Disable boost gate
			ClosePWM1();
			CloseTimer2();
			PSTRCONbits.STRB = 0;		// Disable buck gate PWM
			PSTRCONbits.STRA = 0;		// Disable boost gate PWM
			mode = CHG_OFF;				// Finish
		}
	}
	return(mode);
}

//********************************************************************
//	C h a r g e   c o n t r o l  -  Buck or Through
//********************************************************************
unsigned char chg_buck(unsigned char mode) {
//	static int  duty     = 0;			// PWM duty rate
//	static int  incr     = PWMSTEP_BUCK;// next duty offset
//	static int  ibat_b   = 0;			// before ibat
//	static int  vol_lim  = VBAT_MAX;	//
//	static int  cur_lim  = IBAT_MAX;	//
//	static char climbctr = CLIMBSRCH;	// Up hill counter


	//
	// Fade in buck/through mode, and tracking
	//
	if ((mode == CHG_BUCK) || (mode == CHG_THROUGH)) {

		// PWM module open
		if (!PSTRCONbits.STRB) {		// PWM not set
			OpenTimer2(TIMER_INT_OFF & T2_PS_1_1 & T2_POST_1_1);
			OpenPWM1(PWMPERIOD >> 2);	// Divide 4
			LATCbits.LATC4   = 1;		// Enable buck gate(dummy)
			LATCbits.LATC5   = 0;		// Disable boost gate
			PSTRCONbits.STRA = 0;		// Disconnect boost gate from PWM
			PSTRCONbits.STRB = 1;		// Connect buck gate to PWM
			duty = PWMLOLIM_BUCK;
		}

		// Taget limit
		if (vol_lim >= VBAT_MAX) {		// Max charge mode
			if ((vbat >= VBAT_MAX) && (ibat < IBAT_CHGOFF)) {
				vol_lim = VBAT_TRICKLE;	// End of max charge
				cur_lim = IBAT_TRICKLE;
			}
		} else { 						// Trickle(float) charge mode
			if (vbat < VBAT_DECR) {
				vol_lim = VBAT_MAX;		// Start max charge
				cur_lim = IBAT_MAX;
			}
		}

		// Target seek
		if ((vbat > vol_lim) || (ibat > cur_lim)) {
			duty -= PWMSTEP_BUCK;
			if (duty < PWMLOLIM_BUCK) {
				duty = PWMLOLIM_BUCK;
				mode = CHG_OFF;			// Fail! Force finish
			}
		} else if (mode == CHG_THROUGH) {
			if (duty < PWMHILIM_THROUGH) {
				duty += PWMSTEP_THROUGH;// Slow gate open
				if (duty > PWMHILIM_THROUGH) {	// Full end
					duty = PWMHILIM_THROUGH;
				}
			}
		} else {						// mode == CHG_BUCK
			int d;

			d = ibat - ibat_b;
			if (d < 0) {				// Down? Continue?
				if (--climbctr <= 0) {
					incr = -incr;
					climbctr = CLIMBSRCH;
				}
			} else if (d > 0) {
				climbctr = CLIMBSRCH;
			}
		    duty  += incr;
			ibat_b = ibat;
			if (duty < PWMLOLIM_BUCK) {
				duty = PWMLOLIM_BUCK;
				incr = -incr;
			} else if (duty > PWMHILIM_BUCK) {
				duty = PWMHILIM_BUCK;
				incr = -incr;
			}
		}
		SetDCPWM1(duty);

	//
	// Fade out
	//
	} else {
		if (duty > PWMLOLIM_THROUGH) {
			duty -= PWMSTEP_THROUGH;
			SetDCPWM1(duty);
			mode = CHG_THROUGH;			// Continued
		} else {
			LATCbits.LATC4 = 0;			// Disable buck gate
			LATCbits.LATC5 = 0;			// Disable boost gate
			ClosePWM1();
			CloseTimer2();
			PSTRCONbits.STRB = 0;		// Disable buck gate PWM
			PSTRCONbits.STRA = 0;		// Disable boost gate PWM
			mode = CHG_OFF;				// Finish
		}
	}
	return(mode);
}

//********************************************************************
//	C h a r g e   c o n t r o l  -  Off
//********************************************************************
unsigned char chg_off(unsigned char mode) {
	return(mode);
}

//********************************************************************
//	D i s p l a y
//********************************************************************
void display(unsigned int sec_day, unsigned int sec_night, unsigned char chg_mode) {
	static unsigned int	sec_day_b = 0;	// Newday check
	static long	wssol   = 0;			// [mWsec]
	static long	wsbat   = 0;			// [mWsec]
	static long	wslod   = 0;			// [mWsec]
	static char openflg = 0;			// port status
#define HISTENTRY	10
	static struct {
		int whsol;
		int whbat;
		int whlod;
		int day;
		int night;
	} histtbl[HISTENTRY] = {
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0
	};


	//
	// Daily process
	//
	if (sec_day_b > sec_day) {
		unsigned char i,j;

		for(i = HISTENTRY-1, j = HISTENTRY-2; i > 0; --i, --j) {
			histtbl[i].whsol = histtbl[j].whsol;
			histtbl[i].whbat = histtbl[j].whbat;
			histtbl[i].whlod = histtbl[j].whlod;
			histtbl[i].day   = histtbl[j].day;
			histtbl[i].night = histtbl[j].night;
		}

		wssol = 0;
		wsbat = 0;
		wslod = 0;
	}

	//
	// Summary
	//
	wssol += psol;
	wsbat += pbat;
	wslod += plod;

	//
	// Today's statistics
	//
	histtbl[0].whsol = wssol / 36000;	// [cWh] Wh*10^-2
	histtbl[0].whbat = wsbat / 36000;
	histtbl[0].whlod = wslod / 36000;
	histtbl[0].day   = (long)sec_day   * 5 / 180;	// sec/3600*10000 [ch] h*10^-2
	histtbl[0].night = (long)sec_night * 5 / 180;

	//
	// Output to serial port
	//
	if (PORTCbits.RC7){					// Vbus +5V
		static char buf[10];
#define DISPLAY_ITEM 10
		int  val[DISPLAY_ITEM];
		char i;
		char *p;

		// USB Port open
		if (!openflg) {
			openflg = 1;
			USBDeviceAttach();
		}

		// Statistics
		if (openflg < STAT_DELAY) {
			openflg++;
			if (openflg == STAT_DELAY) {
				char i;

#ifdef CSV_HEAD
				tputrs("Wsol,Wbat,Wlod,Day,Night\r\n");
#endif
				for(i = 0; i < HISTENTRY; i++) {
					char j;

					val[0] = histtbl[i].whsol;
					val[1] = histtbl[i].whbat;
					val[2] = histtbl[i].whlod;
					val[3] = histtbl[i].day;
					val[4] = histtbl[i].night;
					for(j = 0; j < 5; j++) {
						itoanm(val[j], buf, 6, 2);		// ZZ9.99
						buf[6] = ',';
						buf[7] = '\0';
						tputs(buf);
					}
					buf[0] = '\r';
					buf[1] = '\n';
					buf[2] = '\0';
					tputs(buf);
				}
#ifdef CSV_HEAD
				tputrs("Voc,Vsol,Isol,Psol,Vbat,Ibat,Pbat,Ilod,Plod,duty,stat\r\n");
#endif
			}
		}

		// Set value
		val[0]  = vsoloc;
		val[1]  = vsol;
		val[2]  = isol;
		val[3]  = psol;
		val[4]  = vbat;
		val[5]  = ibat;
		val[6]  = pbat;
		val[7]  = ilod;
		val[8]  = plod;
		val[9]  = ((int)chg_mode * 1000) + ((CCPR1L << 2)|((CCP1CON&0x30) >> 4));	// mode.duty

		// Send item
		for(i = 0; i < DISPLAY_ITEM; i++) {
			itoanm(val[i], buf, 7, 3);	// ZZ9.999
			buf[7] = ',';
			buf[8] = '\0';
			tputs(buf);
		}

		// icon
		p = buf;
		*p++ = '"';
		if (sec_day != sec_day_b) {	// Day time
			*p++ = 'S';
		}
		if (LATCbits.LATC4) {		// Solar-cell gate open
			*p++ = 'G';
		}
		if (PSTRCONbits.STRA) {		// Buck/Boost
			*p++ = 'U';
		}
		if (PSTRCONbits.STRB) {
			*p++ = 'D';
		}
		if (LATBbits.LATB7) {		// Load cut
			*p++ = 'L';
		}

		*p++ = '"';
		*p++ = '\r';
		*p++ = '\n';
		*p++ = '\0';
		tputs(buf);

	//
	// Serial port close
	//
	} else {
		if (openflg) {
			openflg = 0;
			USBDeviceDetach();
		}

	}

	sec_day_b = sec_day;

	return;
}

//********************************************************************
//
//		M A I N
//
//********************************************************************
void main(void){
	unsigned int	sec_day;			// sunshine time
	unsigned int	sec_night;			// dark time
	unsigned char	world_stat;			// bright
	unsigned char	world_stat_b;		// Before value
	unsigned char	chg_mode;			// charge mode
	
	//
	// init
	//
	// Config
	ANSEL             = 0x00;			//
	ANSELH            = 0x00;			//
	LATCbits.LATC4    = 0;				//
	TRISCbits.TRISC4  = 0;				// Q1 Gate: buck PWM (Solar-cell input)
	LATCbits.LATC5    = 0;				//
	TRISCbits.TRISC5  = 0;				// Q2 Gate: boost PWM (Coil charge)
	LATBbits.LATB7    = 0;				//
	TRISBbits.TRISB7  = 0;				// Q3 Gate: Load output
	PSTRCONbits.STRA  = 0;				// Disable boost gate PWM
	PSTRCONbits.STRB  = 0;				// Disable buck gate PWM
	WPUA              = 0;				// Internal pullup default all off
	WPUB              = 0;				//
	WPUAbits.WPUA3    = 1;				// SW1 pullup
	INTCON2bits.RABPU = 0;				// Enable pullup
	WDTCONbits.SWDTEN = 1;				// WDT start for about 1sec sleep
	OSCCONbits.IDLEN  = 1;				// No stop peripherals clock on SLEEP
	INTCONbits.GIE    = 1;
	USBDeviceInit();
	OpenTimer0(TIMER_INT_ON & T0_16BIT & T0_SOURCE_INT & T0_PS_1_256);	// 1Hz

	// Work area
	sec_day      = 0;					// 
	sec_night    = 0;					// 
	world_stat   = WORLD_DAY;			// 
	world_stat_b = WORLD_DAY;			// 
	chg_mode     = CHG_OFF;				//

	//
	// 1sec Event loop
	//
	for(;;) {

		//
		// Look
		//
		ClrWdt();
		measure(!(tic & 0x7));						// Answer global area "v*,i*,p*"

		// Valuator
		if (world_stat == WORLD_DAY) {
			if (vsoloc < VOC_SUNSET) {
				world_stat = WORLD_NIGHT;
			}
			sec_day++;
		} else {
			if (vsoloc > VOC_SUNRISE) {
				world_stat = WORLD_DAY;
			}
			sec_night++;
		}
		if (world_stat_b != world_stat) {
			world_stat_b = world_stat;
			if (world_stat == WORLD_DAY) {	// sunrise?
				sec_day = 0;
			} else {
				sec_night = 0;
			}
		}

		//
		// Action
		//
		//loadctrl();
		{
			if (LATBbits.LATB7) {		// Output on
				if (vbat < VBAT_EMPTY) {
					LATBbits.LATB7 = 0;	// load disconnect
				}
			} else {
				if (vbat > VBAT_NORMAL) {
					LATBbits.LATB7 = 1;	// load connect
				}
			}
		}
		//chargectrl();
		{
			unsigned char chg_mode_new;

			// New mode
			chg_mode_new = CHG_OFF;
			if (world_stat == WORLD_DAY) {
				if (vsoloc < VOC_BOOSTON) {
					chg_mode_new = CHG_BOOST;
				} else if (vsoloc > VOC_BUCKON) {
					chg_mode_new = CHG_BUCK;
				} else {
					chg_mode_new = CHG_THROUGH;
				}
				if (!PORTAbits.RA3) {		// SW1 pushed
					chg_mode_new = CHG_THROUGH;
				}
			}

			// Charge action
			switch(chg_mode) {
			case CHG_BOOST:
				chg_mode = chg_boost(chg_mode_new);
				break;
			case CHG_BUCK:
			case CHG_THROUGH:
				chg_mode = chg_buck(chg_mode_new);
				break;
			default:
				chg_mode = chg_off(chg_mode_new);
				break;
			}
		}

		//
		// Show
		//
		display(sec_day, sec_night, chg_mode);

		//
		// Interval wait
		//
		{
			int t;

			t = tic;
			while(t == tic){			// Timer0 intr
				Sleep();
				Nop();
			}
		}
//		Delay10KTCYx(100);
	}
}
